local _, addon = ...
local mod = {}

local _G = _G
local pairs, ipairs = _G.pairs, _G.ipairs
local tinsert, tremove = _G.table.insert, _G.table.remove
local select, type, unpack = _G.select, _G.type, _G.unpack
local setmetatable, getmetatable = _G.setmetatable, _G.getmetatable

-- ==================== table functions ==================== --

-- complete table with items from another
function mod.fillTable(t1, t2)
  for i, v in pairs(t2) do
    if t1[i] == nil then
      t1[i] = v
    elseif type(v) == 'table' then
      mod.fillTable(v, t2[i])
    end
  end
end

-- merges two tables.
function mod.mixTable(t1, t2)
  for i, v in pairs(t2) do
    if t1[i] == nil then
      t1[i] = v
    elseif type(v) == 'table' then
      mod.mixTable(v, t2[i])
    else
      t1[i] = v
    end
  end
end

-- copies anything (int, table...)
-- copied from: http://lua-users.org/wiki/CopyTable
function mod.deepCopy(src)
  local lookup_table = {}
  local function _copy(src)
    if type(src) ~= 'table' then
      return src
    elseif lookup_table[src] then
      return lookup_table[src]
    end
    local new_table = {}
    lookup_table[src] = new_table
    for k, v in pairs(src) do
      new_table[_copy(k)] = _copy(v)
    end
    return setmetatable(new_table, getmetatable(src))
  end

  return _copy(src)
end

-- Checks if the given entry exists in the table:
function mod.checkEntry(t, e)
  for i, v in ipairs(t) do
    if v == e then
      return true -- At least one
    end
  end
  return false
end

-- returns the given table's length
function mod.tableLen(t)
  local length = 0
  for _, _ in pairs(t) do
    length = length + 1
  end
  return length
end

-- checks if the given table is empty
function mod.isEmpty(t)
  for _, _ in pairs(t) do
    return false
  end
  return true
end

-- ==================== strings functions ==================== --

-- better string trimming
function string.trim(str)
  return str:gsub('^%s*(.-)%s*$', '%1')
end

-- ==================== frames functions ==================== --

-- toggle element visibility
function mod.toggle(f, callback)
  if not f then
    return
  elseif f:IsShown() then
    f:Hide()
  else
    f:Show()
  end
  if callback and type(callback) == 'function' then
    callback()
  end
end

-- show/hide frame
function mod.showHide(f, cond, callback)
  if not f then
    return
  elseif cond and not f:IsShown() then
    f:Show()
  elseif not cond and f:IsShown() then
    f:Hide()
  end
  if callback and type(callback) == 'function' then
    callback()
  end
end

-- shows with callback
function mod.show(f, callback)
  if f and not f:IsShown() then
    f:Show()
    if callback and type(callback) == 'function' then
      callback()
    end
  end
end

-- conditional frame show
function mod.showIf(f, cond, callback)
  if f and not f:IsShown() and cond then
    f:Show()
    if callback and type(callback) == 'function' then
      callback()
    end
  end
end

-- hide frame
function mod.hide(f, callback)
  if f and f:IsShown() then
    f:Hide()
    if callback and type(callback) == 'function' then
      callback()
    end
  end
end

-- conditional frame hide
function mod.hideIf(f, cond, callback)
  if f and f:IsShown() and cond then
    f:Hide()
    if callback and type(callback) == 'function' then
      callback()
    end
  end
end

-- enable/disable button
function mod.enableDisable(b, cond)
  if not b then
    return
  elseif cond and b:IsEnabled() == 0 then
    b:Enable()
  elseif not cond and b:IsEnabled() == 1 then
    b:Disable()
  end
end

-- toggle button highlight status
function mod.highlight(b, cond)
  if not b then
    return
  elseif cond then
    b:LockHighlight()
  else
    b:UnlockHighlight()
  end
end

-- change element text
function mod.setText(f, str1, str2, cond)
  if not f then
    return
  elseif cond then
    f:SetText(str1)
  else
    f:SetText(str2)
  end
end

-- conditional texture change
function mod.setTexture(f, txr1, txr2, cond)
  if not f then
    return
  elseif cond then
    f:SetTexture(txr1)
  else
    f:SetTexture(txr2)
  end
end

-- conditional texture change
function mod.setNormalTexture(f, txr1, txr2, cond)
  if not f then
    return
  elseif cond then
    f:SetNormalTexture(txr1)
  else
    f:SetNormalTexture(txr2)
  end
end

-- better frame onupdate
function mod.update(f, name, period, elapsed)
  local t = name and f[name] or 0
  t = t + elapsed
  if t > period then
    f[name] = 0
    return true
  end
  f[name] = t
  return false
end

-- ==================== task manager ==================== --
do
  local GetTime = _G.GetTime

  -- Table of scheduled tasks:
  local tasks = {}

  -- Schedule a task:
  function mod.schedule(sec, func, ...)
    local task = {}
    task.time = GetTime() + sec
    task.func = func
    task.args = {}
    for i = 1, select('#', ...) do
      task.args[i] = select(i, ...)
    end
    -- tasks[#tasks+1] = task
    tinsert(tasks, task)
  end

  -- Reschedule a task:
  function mod.reschedule(func, sec)
    for i = 1, #tasks do
      if tasks[i].func == func then
        tasks[i].time = tasks[i].time + sec
        break
      end
    end
  end

  -- Unschedule a task:
  function mod.unschedule(func)
    for i, v in pairs(tasks) do
      if func == v.func then
        tremove(tasks, i)
        break
      end
    end
  end

  -- Run all scheduled tasks:
  function mod.run()
    local now = GetTime()
    for i = 1, #tasks do
      local task = tasks[i]
      if task and type(task.func) == 'function' and task.time <= now then
        task.func(unpack(task.args, 0))
        tremove(tasks, i) -- Only once!
      end
    end
  end
end

-- ==================== chat output functions ==================== --
do
  local CHAT_FRAME = _G.DEFAULT_CHAT_FRAME
  local SendChatMessage = _G.SendChatMessage
  local tconcat = _G.table.concat

  local temp = {}

  local function print_output(frame, ...)
    local n = 1
    if not temp[n] then temp[n] = '|cff33ff99KTracker|r:' end
    for i = 1, select('#', ...) do
      n = n + 1
      temp[n] = tostring(select(i, ...))
    end
    frame:AddMessage(tconcat(temp, ' ', 1, n))
  end

  function mod.print(...)
    local frame = ...
    if type(frame) == 'table' and frame.AddMessage then
      return print_output(frame, select(1, ...))
    else
      return print_output(CHAT_FRAME, ...)
    end
  end

  function mod.printf(...)
    local frame = ...
    if type(frame) == 'table' and frame.AddMessage then
      return print_output(frame, format(select(1, ...)))
    else
      return print_output(CHAT_FRAME, format(...))
    end
  end

  -- Prints a green success message.
  function mod.printSuccess(text)
    text = '|cff95dd93'..tostring(text)..'|r'
    return mod.print(text)
  end

  -- Prints an orange warning message.
  function mod.printWarning(text)
    text = '|cffff8a00'..tostring(text)..'|r'
    return mod.print(text)
  end

  -- Prints a red error message.
  function mod.printError(text)
    text = '|cffff3c00'..tostring(text)..'|r'
    return mod.print(text)
  end

  -- add functions to addon scope
  function addon:Print(...) return mod.print(...) end
  function addon:PrintSuccess(...) return mod.printSuccess(...) end
  function addon:PrintWarning(...) return mod.printWarning(...) end
  function addon:PrintError(...) return mod.printError(...) end
end

-- ==================== tooltip functions ==================== --
do
  local color = _G.HIGHLIGHT_FONT_COLOR
  local GameTooltip_SetDefaultAnchor = _G.GameTooltip_SetDefaultAnchor

  -- Show the tooltip
  -- @param   obj   the frame data
  local function showTooltip(frame)
    -- Is the anchor manually set?
    if not frame.tooltip_anchor then
      GameTooltip_SetDefaultAnchor(KTrackerTooltip, frame)
    else
      KTrackerTooltip:SetOwner(frame, frame.tooltip_anchor)
    end

    KTrackerTooltip:ClearLines()

    if frame.tooltip_title then
      KTrackerTooltip:SetText(frame.tooltip_title)
    end

    if frame.tooltip_text then
      if type(frame.tooltip_text) == 'string' then
        KTrackerTooltip:AddLine(frame.tooltip_text, color.r, color.g, color.b, true)
      elseif type(frame.tooltip_text) == 'table' then
        for _, l in ipairs(frame.tooltip_text) do
          KTrackerTooltip:AddLine(l, color.r, color.g, color.b, true)
        end
      end
    end

    if frame.tooltip_item then
      KTrackerTooltip:SetHyperlink(frame.tooltip_item)
    end

    KTrackerTooltip:Show()
  end

  -- Hides the tooltip:
  local function hideTooltip()
    KTrackerTooltip:Hide()
  end

  -- Sets frame OnEnter and OnLeave scripts.
  -- @param   frame the frame to attach tooltip to
  -- @param   text  the text to display (string or table)
  -- @param   anchor  where to display the tooltip
  -- @param   title   whether to give the tooltip a title
  function mod.setTooltip(frame, text, anchor, title)
    if frame then
      -- Prepare the text
      frame.tooltip_text = text and text or frame.tooltip_text
      frame.tooltip_anchor = anchor and anchor or frame.tooltip_anchor
      frame.tooltip_title = title and title or frame.tooltip_title
      -- No title or text? nothing to do...
      if frame.tooltip_title or frame.tooltip_text or frame.tooltip_item then
        frame:SetScript('OnEnter', showTooltip)
        frame:SetScript('OnLeave', hideTooltip)
      end
    end
  end

  -- Remove all previously register tooltip script.
  function mod.unsetTooltip(frame)
    if frame then
      frame:SetScript('OnEnter', nil)
      frame:SetScript('OnLeave', nil)
    end
  end
end

-- ==================== callbacks functions ==================== --
do
  local pcall, error = _G.pcall, _G.error

  -- Tables of registered callbacks:
  local callbacks = {}

  -- Register a callback function:
  function mod.registerCallback(e, func)
    if e and type(func) == 'function' then
      callbacks[e] = callbacks[e] or {}
      tinsert(callbacks[e], func)
      return #callbacks
    end
    error('Usage: mod.registerCallback(action, callback)', 2)
  end

  -- Executes a previously register action.
  function mod.triggerEvent(e, ...)
    if callbacks[e] then
      for i, v in ipairs(callbacks[e]) do
        local ok, err = pcall(v, a, ...)
        if not ok then
          mod.print(format('Error while executing callback %s for event %s: %s', tostring(v), tostring(a), err))
        end
      end
    end
  end
end

-- ===================== unit functions ==================== --

do
  local GetNumPartyMembers = _G.GetNumPartyMembers
  local GetNumRaidMembers = _G.GetNumRaidMembers
  local GetRaidRosterInfo = _G.GetRaidRosterInfo
  local UnitName, UnitGUID = _G.UnitName, _G.UnitGUID

  -- Checks if the player is in a party group:
  -- Checks if the player is in a party group:
  function mod.isInParty()
    return (GetNumPartyMembers() > 0)
  end

  -- Checks if the player is in a party group:
  function mod.isInRaid()
    return (GetNumRaidMembers() > 0)
  end

  -- Returns the player's UnitID
  function mod.getPlayerUnitId(playerName)
    playerName = playerName or UnitName('player')
    local unitId = 'none'
    local uId = mod.isInRaid() and 'raid' or 'party'
    local count = math.max(GetNumRaidMembers(), GetNumPartyMembers())
    for i = 0, count do
      local id = (i == 0 and 'player') or uId..i
      if playerName == UnitName(id) then
        unitId = id
        break
      end
    end
    return unitId
  end
end

-- ================ class colors ================ --

do
  -- Cache colors:
  local colors = CUSTOM_CLASS_COLORS or RAID_CLASS_COLORS

  -- RGB2Hex Conversion:
  function mod.RGB2HEX(r, g, b)
    return format('%02x%02x%02x', r * 255, g * 255, b * 255)
  end

  -- Returns R, G, B on the given class:
  function mod.classColor(name)
    name = (name == 'DEATH KNIGHT') and 'DEATHKNIGHT' or name
    if not colors[name] then return 1, 1, 1 end
    return colors[name].r, colors[name].g, colors[name].b
  end

  -- Returns HEX class color:
  function mod.classColorHex(name)
    local r, g, b = mod.classColor(name)
    return mod.RGB2HEX(r, g, b)
  end
end

-- ================ synchronization functions ================ --

do
  local IsInInstance = _G.IsInInstance
  local GetRealNumRaidMembers = _G.GetRealNumRaidMembers
  local GetRealNumPartyMembers = _G.GetRealNumPartyMembers

  -- using ChatThrottleLib
  local SendAddonMessage = _G.SendAddonMessage
  if ChatThrottleLib and ChatThrottleLib.SendAddonMessage then
    SendAddonMessage = function(...)
      ChatThrottleLib:SendAddonMessage('NORMAL', ...)
    end
  end

  -- Sends an addOn message to the appropriate channel:
  function mod.sync(prefix, msg, channel, target)
    if not channel then
      local zone = select(2, IsInInstance())
      if zone == 'pvp' or zone == 'arena' then
        channel = 'BATTLEGROUND'
      elseif GetRealNumRaidMembers() > 0 then
        channel = 'RAID'
      elseif GetRealNumPartyMembers() > 0 then
        channel = 'PARTY'
      end
    end
    if channel then SendAddonMessage(prefix, msg, channel, target) end
  end
end

-- ==================== classes functions ==================== --
--
-- this part was copied from Classy-1.0.lua
-- all credits to its owner
--

mod.Class = {}
function mod.Class:New(from, parent)
  local class = CreateFrame(from)
  class.mt = {__index = class}

  if parent then
    class = setmetatable(class, {__index = parent})

    class.super = function(self, method, ...)
      parent[method](self, ...)
    end
  end

  class.Bind = function(self, obj)
    return setmetatable(obj, self.mt)
  end

  return class
end

addon.utils = mod
